"""This is ndxtest. It lets you test trading strategies on the S&P 500 index.

Refer to https://github.com/lksdnx/ndxtest and https://ndxtest.readthedocs.io/en/latest/
for information on how to use this package.

For information on the usage of this package please refer to the online documentation.

Modules
_______
    backtest.py
    utils.py
    indicators.py
"""

import backtest
import indicators
import utils

